#!/usr/bin/python
import subprocess
import string
import os
import os.path
import sys
import ConfigParser

abs_path = os.path.abspath(sys.argv[0])
basename = os.path.basename(sys.argv[0])
dirname = abs_path.replace(basename, "")
if "diskmon/bin" in dirname:
   if not "%slib" % dirname[:-4] in sys.path:
      sys.path.insert(0,"%slib" % dirname[:-4]) # in case you run this in benchmark directory as diskmon/bin/ana_snmptrap_sender.py --anaDisks e.g.
if not "%s/diskmon/lib" % dirname in sys.path:
    sys.path.insert(0,"%sdiskmon/lib" % dirname) # always with regard to parent script
if not "/usr/bp/lib/monitoring" in sys.path:
    sys.path.append("/usr/bp/lib/monitoring")
    
from bpdb     import *   # include config << bpdb as well
from optparse import OptionParser

parser = OptionParser()
parser.add_option("-d", "--debug",   dest="debug", default=False, action='store_true',
                  help="debug option to show detail commands to test")
parser.add_option("-f", "--force", dest="force", default=False, action='store_true',
                  help="sending trap ignoring pass threshold (Sunday midnight)")
parser.add_option("-c", "--confirm",   dest="confirm", default=False, action='store_true',
                  help="Without confirm flag, snmp won't be sent")
(opts, args) = parser.parse_args()

CONFIG.debug            =opts.debug
CONFIG.confirm          =opts.confirm
CONFIG.force            =opts.force

from trap   import *   # include config << bpdb as well
from config import get_master_ini,get_first_word,get_printout
from util   import get_queryResult, binary_bit_lifter, get_hex, does_exist

CONFIG.db_accessible =True
CONFIG.dbOff_reason=""
_master_ini ="/usr/bp/bpinit/master.ini"

if opts.debug:
   _master_ini ="/var/tmp/master.ini"

import datetime
_now=("%s"%datetime.datetime.now()).split(".")[0]

def is_db_accessible():
   reason=""
   import commands      
   lines = commands.getoutput("/bin/ps aux | /bin/grep postgres | /bin/grep data").splitlines()
   dbON     =True
   # psql: could not fork new process for connection: Cannot allocate memory
   # True 1 ['root      9314  0.0  0.0  63848  1048 pts/2    S+   09:34   0:00 sh -c { ps aux | grep postgres | grep data; } 2>&1']

   lcnt = 0
   for line in lines:
      if "grep" in line:       continue
      lcnt +=1
   if not lcnt:
      dbON = False
      reason="no process postgres running"
      return reason
   
   # without bpdb postgres, you will see the following message from crontab:
   # psql: FATAL:  role "root" does not exist
   lines= commands.getoutput("/usr/bin/psql bpdb postgres -c \"select * from bp.nodes limit 1\"").splitlines()
   for line in lines:
      if "could not access status of transaction" in line or \
         "timeout expired" in line or \
         "Connection refused" in line or \
         "the database system is starting up" in line or \
         "too many clients" in line:
         dbON = False
         reason=line[:-1] # getting rid of return character (\n)
         return reason
      
   # Do not run while VACUUM is running
   lines = commands.getoutput("/bin/ps aux | /bin/grep VACUUM | /bin/grep -v grep").splitlines()
   if len(lines):
      dbON=False
      line=lines[0]
      reason=line[:-1]
      
   return reason

def main():

   CONFIG.dbOff_reason=is_db_accessible()   
   CONFIG.db_accessible = not len(CONFIG.dbOff_reason)

   shortName="ProactiveDetection"
   
   import psycopg2
   dbname,username='bpdb','postgres'
   con=None

   try:
      con =psycopg2.connect(database=dbname, user=username)
      
      notification_id = 222
      category_id     = 4 # Backup
      source_id       =10 # backup
      severity        = 2 #WARNING
      target          = "aTypicalChangeRate"
      message =ransomWareDetect(con,shortName)

      if len(message):
         print "%s"%_now, message
      # THIS IS ONE OFF RELEASE ASSUMING NO UPGRADE OF UNITRENDS-RR
      AlertPost(con,notification_id,category_id,source_id,severity,target,message)

   except psycopg2.DatabaseError, e:
      print 'Error %s' %e
      if con:
         con.close()
      sys.exit(1)
      
   finally:
      if con:
         con.close()
       
   return
    
def AlertPost(con,notification_id,category_id,source_id,severity,target,message):
   cur =con.cursor()
   query="""
   SELECT notification_id FROM bp.notifications 
   WHERE notification_id=%d
   """ % notification_id
   cur.execute(query)
   doesNotificationExist=len(cur.fetchall())
   
   query ="""
   SELECT alert_id
   FROM bp.alerts
   WHERE target='aTypicalChangeRate'
       AND closed=FALSE;
   """
   cur.execute(query)
   if opts.debug: print query
   openAlerts=cur.fetchall()

   query=""
   import time
   epoch_time=int(time.time())

   # If atypical change rate is noticed
   if len(message):    
      # if there is alert posted
      if len(openAlerts):
         alert_id= (openAlerts[0])[0]
         query="""
         UPDATE bp.alerts SET updated=%d,alert_text='%s' WHERE alert_id=%d
         """ % (epoch_time,message,alert_id)
      else:
         # if notification exists, then use ids 
         if doesNotificationExist: 
            query="""
            INSERT INTO bp.alerts
            (alert_text, source_id, target, severity, created, updated, notification_id)
            VALUES
            ('%s',       %d,        '%s',   %d,       %d,      %d,      %d)
            """ % (message, source_id, target, severity, epoch_time, epoch_time, notification_id)
         else:
            query="""
            INSERT INTO bp.alerts
            (alert_text, source_id, target, severity, created, updated)
            VALUES
            ('%s',       %d,        '%s',   %d,       %d,      %d)
            """ % (message, source_id, target, severity, epoch_time, epoch_time)
   # message could be empty by filtering or resolved
   else:
      if len(openAlerts):
         alert_id= (openAlerts[0])[0]
         query="""
         UPDATE bp.alerts SET updated=%d,closed=TRUE WHERE alert_id=%d
         """ % (epoch_time,alert_id)

   if opts.confirm:
      if len(query):
         cur.execute(query)
         con.commit()
   else:
      print query
      
    

def getExclusionList(shortName):
   exclusionListOfVMwareVMs,exclusionListOfHypervVMs,exclusionListOfXenVMs,exclusionListOfNodes=[],[],[],[]
   excluded_vmware_vms_query_list,excluded_hyperv_vms_query_list,excluded_xen_vms_query_list,excluded_nodes_query_list="","","",""
   vmware_vms   =get_master_ini(shortName,"exclusion_list_of_vmware_vm_names")
   hyperv_vms   =get_master_ini(shortName,"exclusion_list_of_hyperv_vm_names")
   xen_vms   =get_master_ini(shortName,"exclusion_list_of_xen_vm_names")
   nodes        =get_master_ini(shortName,"exclusion_list_of_node_names")

   print "%s" %_now, "ExclusionList: nodes: %s, vmware_vms: %s, hyperv_vms: %s, xen_vms: %s" %(nodes,vmware_vms,hyperv_vms,xen_vms)
   
   if vmware_vms !="ERROR" and len(vmware_vms):
      for vm in vmware_vms.split(","):
         exclusionListOfVMwareVMs.append(vm)
   if hyperv_vms !="ERROR" and len(hyperv_vms):
      for vm in hyperv_vms.split(","):
         exclusionListOfHypervVMs.append(vm)
   if xen_vms !="ERROR" and len(xen_vms):
      for vm in xen_vms.split(","):
         exclusionListOfXenVMs.append(vm)
   if nodes !="ERROR" and len(nodes):
      for node in nodes.split(","):
         exclusionListOfNodes.append(node)

   if len(exclusionListOfVMwareVMs):
      if opts.debug: print exclusionListOfVMwareVMs
      excluded_vmware_vms_query_list   =get_query_list(exclusionListOfVMwareVMs)
   if len(exclusionListOfHypervVMs):
      if opts.debug: print exclusionListOfHypervVMs
      excluded_hyperv_vms_query_list   =get_query_list(exclusionListOfHypervVMs)
   if len(exclusionListOfXenVMs):
      if opts.debug: print exclusionListOfXenVMs
      excluded_xen_vms_query_list   =get_query_list(exclusionListOfXenVMs)
   if len(exclusionListOfNodes):
      if opts.debug: print exclusionListOfNodes
      excluded_nodes_query_list =get_query_list(exclusionListOfNodes)
   
   return (excluded_vmware_vms_query_list,excluded_hyperv_vms_query_list,excluded_xen_vms_query_list,excluded_nodes_query_list)


def getExclusionInstanceIdList(con,shortName):
   
   cur =con.cursor()
   
   (excluded_vmware_vms_query_list,excluded_hyperv_vms_query_list,excluded_xen_vms_query_list,excluded_nodes_query_list)=getExclusionList(shortName)

   ids=[]

   if len(excluded_vmware_vms_query_list):
      query="""
      SELECT instance_id
      FROM bp.application_instances
      WHERE app_id=40 AND key2 in %s
      """ % excluded_vmware_vms_query_list

      if opts.debug: print query
   
      cur.execute(query)
      rows=cur.fetchall()
      for i in range(len(rows)):
         row=rows[i]
         ids.append(row[0])


   if len(excluded_hyperv_vms_query_list):
      query="""
      SELECT instance_id
      FROM bp.application_instances
      WHERE app_id>=50 AND app_id<60 AND key1 in %s
      """ % excluded_hyperv_vms_query_list

      if opts.debug: print query
      
      cur.execute(query)
      rows=cur.fetchall()
      for i in range(len(rows)):
         row=rows[i]
         ids.append(row[0])

   if len(excluded_xen_vms_query_list):
      query="""
      SELECT instance_id
      FROM bp.application_instances
      WHERE app_id>=140 AND app_id<150 AND key2 in %s
      """ % excluded_xen_vms_query_list

      if opts.debug: print query
      
      cur.execute(query)
      rows=cur.fetchall()
      for i in range(len(rows)):
         row=rows[i]
         ids.append(row[0])
         

   if len(excluded_nodes_query_list):
      # file-level application only
      query="""
      SELECT instance_id
      FROM   bp.nodes AS n, bp.application_instances AS a
      WHERE  n.node_no=a.node_no AND app_id=1 AND node_name <> 'vCenter-RRC' AND node_name in %s;
      """ % excluded_nodes_query_list

      if opts.debug: print query

      cur.execute(query)
      rows=cur.fetchall()
      for i in range(len(rows)):
         row=rows[i]
         ids.append(row[0])

   if opts.debug: print "ids:",ids
   return get_query_list(ids)
   

   
def ransomWareDetect(con, shortName):

   hours=get_master_ini(shortName,"last_time_window_hours")
   if "ERROR" == hours or not len(hours):
      hours=48
   else:
      hours=int(hours)

   seconds=3600*hours   # 172800: 24 hrs

   
   threshold=get_master_ini(shortName,"threshold_percentage_compared_to_avg_change")
   
   if "ERROR" == threshold or not len(threshold):
      threshold=500
   else:
      threshold=int(threshold)
      
   ret ="The predictive analytics engine has detected anomalies on the following systems which probabilistically "
   ret+="matches the behavior of systems impacted by Ransomware. It is recommended to check the systems for malware "
   ret+="and recover them to the latest available backup, if malware is confirmed. Systems :"
   
   #WHERE  status=3168 AND type in (1,2,3,12,13,14,1031,1032,1033)
   # at least 100MB size of first backup

   exclusion_instance_ids=getExclusionInstanceIdList(con,shortName)
      
   extra=""
   if len(exclusion_instance_ids):
      extra="AND instance_id NOT IN %s" %exclusion_instance_ids
         
   query="""
   SELECT * FROM (
       SELECT
           second_backup AS backup_no,
           COUNT(1) AS historical_backups,
           MAX(hist_time) - MIN(hist_time) as hist_time_delta,
           MAX(instance_id) AS instance_id,
           MAX(new_data) / GREATEST( MAX(test_delta_time), 1) AS latest_rate,
           SUM(new_blocks_size) / GREATEST ( MAX(hist_time) - MIN(hist_time), 1) AS historical_rate,
           STDDEV(new_blocks_size) / GREATEST ( MAX(hist_time) - MIN(hist_time), 1) AS historical_rate_dev,
           100 * (MAX(new_data) / GREATEST( MAX(test_delta_time), 1) )
                  / GREATEST( ( sum(new_blocks_size)/ GREATEST( MAX(hist_time) - MIN(hist_time), 1) ),1 ) AS percent_of_historical
       FROM (
             SELECT
                 second_backup,
                 original_backup,
                 hist.backup_no,
                 hist.new_blocks_size,
                 Extract(epoch FROM hist.stat_time) AS hist_time,
                 second_stat_time AS second_backup_end_time,
                 second_new_blocks_size AS new_data,
                 Extract(epoch FROM ( second_stat_time - first_stat_time )) AS test_delta_time,
                 pairs.instance_id
             FROM (
                   SELECT
                       Min(first.backup_no) AS original_backup,
                       Max(first.backup_no) AS first_backup,
                       Max(first.stat_time) AS first_stat_time,
                       second.backup_no AS second_backup,
                       Max(second.new_blocks_size) AS second_new_blocks_size,
                       Max(second.stat_time) AS second_stat_time,
                       Max(second.instance_id) AS instance_id
                   FROM (
                         SELECT
                             b.backup_no,
                             instance_id,
                             stat_time
                         FROM bp.backups b
                             JOIN bp.backup_dedup_stats bds
                             ON b.backup_no = bds.backup_no
                         WHERE status = 3168
                             AND operation = 'inline_dedup'
                             AND type in (1,2,3,12,13,14,1031,1032,1033,4101,4102)
                         ) AS first
                         JOIN
                         (
                          SELECT
                              b.backup_no,
                              instance_id,
                              stat_time,
                              new_blocks_size
                          FROM bp.backups b
                              JOIN bp.backup_dedup_stats bds
                              ON b.backup_no = bds.backup_no
                          WHERE status = 3168
                              AND type in (1,2,3,12,13,14,1031,1032,1033,4101,4102)
                              AND operation = 'inline_dedup'
                              AND ( Extract(epoch FROM Now()) - start_time ) < %d
                          ) AS second
                         ON first.instance_id = second.instance_id
                         WHERE first.backup_no < second.backup_no
                         GROUP  BY second.backup_no
                  ) AS pairs
                  JOIN
                  (
                   SELECT
                       b.backup_no,
                       instance_id,
                       new_blocks_size,
                       operation,
                       stat_time
                   FROM bp.backup_dedup_stats bds
                       JOIN bp.backups b
                       ON bds.backup_no = b.backup_no
                   WHERE status=3168
                       AND type in (1,2,3,12,13,14,1031,1032,1033,4101,4102)
                       AND operation = 'inline_dedup'
                  ) AS hist
                  ON pairs.instance_id = hist.instance_id
                  WHERE  hist.backup_no <= first_backup
                  AND backup_no > original_backup
             ) AS long_data
             GROUP BY second_backup
   ) AS change_rates
   WHERE percent_of_historical > %d
       %s
   ORDER BY instance_id,backup_no
   """ %(seconds,threshold, extra)
   
   if opts.debug:
      print query

   cur =con.cursor()
   cur.execute(query)
   rows=cur.fetchall()
   candidate_ids=[]
   for i in range(len(rows)):
      row=rows[i]
      (second_backup, historical_backups, hist_time_delta,
       instance_id, latest_rate, historical_rate, historical_rate_dev,percent_of_historical)=row

      print "%s"%_now, "instId=%5d, bkno=%5d, nBkups=%3d, hist_days=%3d, last_rate= %4.1fGB/day, avg_rate= %4.1fGB/day, perc_of_historical=%4d%%"\
         %(instance_id,second_backup,historical_backups,int(hist_time_delta/3600./24. +0.5),
           int(10*3600*24*latest_rate/1024./1024./1024.+0.5)/10.,int(10*3600*24*historical_rate/1024./1024./1024.+0.5)/10.,percent_of_historical)
      #if change_rate_perc>=threshold:
      candidate_ids.append(instance_id)

   if len(candidate_ids):
      query="""
      SELECT n.node_no,n.os,instance_id,l.type,node_name,key1,key2
      FROM   bp.nodes as n, bp.application_instances as a, bp.application_lookup as l
      WHERE  n.node_no=a.node_no AND l.app_id=a.app_id AND instance_id in %s
      ORDER BY l.app_id;
      """ % get_query_list(candidate_ids)

      if opts.debug:
         print query
      
      cur.execute(query)
      rows=cur.fetchall()
      #candidates=[]
      for i in range(len(rows)):
         row=rows[i]
         (node_no,os,instance_id,appType,node_name,key1,key2)=row

         theName=node_name  # appType: file-level
         if appType=="Hyper-V":
            theName=key1
            ret+="%s:%s:%s; " %(appType,node_name,theName)
         elif appType in ("VMware","Xen"):
            theName=key2
            ret+="%s:%s:%s; " %(appType,node_name,theName)
         else:   
            ret+="%s; " %(theName)
            
         #candidates.append( (appType,theName) )
      ret+="For more information please refer to KB article https://support.unitrends.com/UnitrendsBackup/s/article/000005442"
   else:
      ret=""
         
      
   
   return ret


def get_query_list(ids):

   strings=""
   if not len(ids):
      return strings
   for id in ids:
      if isinstance(id,str):
         strings+="'%s'," % id
      elif isinstance(id,int):
         strings+="%d," % id

   strings="(%s)" %strings[:-1]
   if opts.debug:
      print strings
   return strings

def get_master_ini(section, para):
   command="/usr/bp/bin/bputil -g %s %s ERROR %s" % (section, para, _master_ini)
   ret=""
   lines=get_printout(command)
   if len(lines):
      ret=lines[0]
   return ret

def set_master_ini(section, para, value):
    run_command( "/usr/bp/bin/bputil -p %s %s %s %s" % (section, para, value, _master_ini), 0, True)


if __name__ == '__main__':
    main()


